# Python Modules
import os

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError
from outsystems.exceptions.environment_not_found import EnvironmentNotFoundError
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
from outsystems.exceptions.app_does_not_exist import AppDoesNotExistError
from outsystems.exceptions.server_error import ServerError
# Functions
from outsystems.lifetime.lifetime_base import send_get_request, send_post_request
from outsystems.lifetime.lifetime_applications import _get_application_info
from outsystems.file_helpers.file import load_data, store_data, clear_cache
# Variables
from outsystems.vars.lifetime_vars import ENVIRONMENTS_ENDPOINT, ENVIRONMENT_APPLICATIONS_ENDPOINT, ENVIRONMENTS_SUCCESS_CODE, \
    ENVIRONMENTS_NOT_FOUND_CODE, ENVIRONMENTS_FAILED_CODE, ENVIRONMENT_APP_SUCCESS_CODE, ENVIRONMENT_APP_NOT_STATUS_CODE, \
    ENVIRONMENT_APP_NO_PERMISSION_CODE, ENVIRONMENT_APP_NOT_FOUND, ENVIRONMENT_APP_FAILED_CODE, ENVIRONMENT_DEPLOYMENT_ZONES_ENDPOINT, \
    ENVIRONMENT_ZONES_SUCCESS_CODE, ENVIRONMENT_ZONES_NOT_STATUS_CODE, ENVIRONMENT_ZONES_NO_PERMISSION_CODE, ENVIRONMENT_ZONES_NOT_FOUND, \
    ENVIRONMENT_ZONES_FAILED_CODE, ENVIRONMENT_APPLICATIONS_SOURCECODE_ENDPOINT, ENVIRONMENT_SOURCECODE_PACKAGE_SUCCESS_CODE, \
    ENVIRONMENT_SOURCECODE_LINK_SUCCESS_CODE, ENVIRONMENT_SOURCECODE_FAILED_CODE
from outsystems.vars.file_vars import ENVIRONMENTS_FILE, ENVIRONMENT_FOLDER, ENVIRONMENT_APPLICATION_FILE, \
    ENVIRONMENT_DEPLOYMENT_ZONES_FILE, ENVIRONMENT_SOURCECODE_FOLDER, ENVIRONMENT_SOURCECODE_STATUS_FILE, \
    ENVIRONMENT_SOURCECODE_LINK_FILE


# Lists all the environments in the infrastructure.
def get_environments(artifact_dir: str, endpoint: str, auth_token: str):
    # Sends the request
    response = send_get_request(
        endpoint, auth_token, ENVIRONMENTS_ENDPOINT, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENTS_SUCCESS_CODE:
        # Stores the result
        store_data(artifact_dir, ENVIRONMENTS_FILE, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENTS_NOT_FOUND_CODE:
        raise EnvironmentNotFoundError(
            "No environments found. Details {}".format(response["response"]))
    elif status_code == ENVIRONMENTS_FAILED_CODE:
        raise ServerError(
            "Failed to list the environments. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the environment key.
def get_environment_key(artifact_dir: str, endpoint: str, auth_token: str, environment_name: str):
    return _find_environment_key(artifact_dir, endpoint, auth_token, environment_name)


# Returns information about the running version of the specified application in a given environment.
def get_environment_app_version(artifact_dir: str, endpoint: str, auth_token: str, extra_data: bool, **kwargs):
    # Tuple with (AppName, AppKey): app_tuple[0] = AppName; app_tuple[1] = AppKey
    app_tuple = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Tuple with (EnvName, EnvKey): env_tuple[0] = EnvName; env_tuple[1] = EnvKey
    env_tuple = _get_environment_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Builds the query and arguments for the call to the API
    query = "{}/{}/{}/{}".format(ENVIRONMENTS_ENDPOINT,
                                 env_tuple[1], ENVIRONMENT_APPLICATIONS_ENDPOINT, app_tuple[1])
    arguments = {"IncludeEnvStatus": extra_data, "IncludeModules": extra_data}
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, arguments)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_APP_SUCCESS_CODE:
        # Stores the result
        filename = "{}.{}{}".format(
            env_tuple[0], app_tuple[0], ENVIRONMENT_APPLICATION_FILE)
        filename = os.path.join(ENVIRONMENT_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENT_APP_NOT_STATUS_CODE:
        raise InvalidParametersError("Error in the request parameters. Params: {}. Details {}".format(
            arguments, response["response"]))
    elif status_code == ENVIRONMENT_APP_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the application in that environment. Details: {}".format(response["response"]))
    elif status_code == ENVIRONMENT_APP_NOT_FOUND:
        raise AppDoesNotExistError(
            "The application does not exist in the environment. Details: {}".format(response["response"]))
    elif status_code == ENVIRONMENT_APP_FAILED_CODE:
        raise ServerError("Failed to access the running version of an application. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns information about the existing deployment zones in a given environment.
def get_environment_deployment_zones(artifact_dir: str, endpoint: str, auth_token: str, **kwargs):
    # Tuple with (EnvName, EnvKey): env_tuple[0] = EnvName; env_tuple[1] = EnvKey
    env_tuple = _get_environment_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Builds the query and arguments for the call to the API
    query = "{}/{}/{}".format(ENVIRONMENTS_ENDPOINT, env_tuple[1], ENVIRONMENT_DEPLOYMENT_ZONES_ENDPOINT)
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_ZONES_SUCCESS_CODE:
        # Stores the result
        filename = "{}.{}".format(
            env_tuple[0], ENVIRONMENT_DEPLOYMENT_ZONES_FILE)
        filename = os.path.join(ENVIRONMENT_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENT_ZONES_NOT_STATUS_CODE:
        raise InvalidParametersError(
            "Failed to access the deployment zones of environment. Details {}".format(response["response"]))
    elif status_code == ENVIRONMENT_ZONES_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to retrieve the deployment zones in that environment. Details: {}".format(response["response"]))
    elif status_code == ENVIRONMENT_ZONES_NOT_FOUND:
        raise EnvironmentNotFoundError(
            "No environment found to retrieve the deployment zones. Details {}".format(response["response"]))
    elif status_code == ENVIRONMENT_ZONES_FAILED_CODE:
        raise ServerError(
            "Failed to access the deployment zones of environment. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the package key to download the source code of the specified application in a given environment.
def get_environment_app_source_code(artifact_dir: str, endpoint: str, auth_token: str, **kwargs):
    # Tuple with (AppName, AppKey): app_tuple[0] = AppName; app_tuple[1] = AppKey
    app_tuple = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Tuple with (EnvName, EnvKey): env_tuple[0] = EnvName; env_tuple[1] = EnvKey
    env_tuple = _get_environment_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Builds the query and arguments for the call to the API
    query = "{}/{}/{}/{}/{}".format(ENVIRONMENTS_ENDPOINT, env_tuple[1],
                                    ENVIRONMENT_APPLICATIONS_ENDPOINT, app_tuple[1],
                                    ENVIRONMENT_APPLICATIONS_SOURCECODE_ENDPOINT)
    # Sends the request
    response = send_post_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOURCECODE_PACKAGE_SUCCESS_CODE:
        return response["response"]
    elif status_code == ENVIRONMENT_SOURCECODE_FAILED_CODE:
        raise ServerError("Failed to access the source code of an application. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns current status of source code package of the specified application in a given environment.
def get_environment_app_source_code_status(artifact_dir: str, endpoint: str, auth_token: str, **kwargs):
    # Tuple with (AppName, AppKey): app_tuple[0] = AppName; app_tuple[1] = AppKey
    app_tuple = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Tuple with (EnvName, EnvKey): env_tuple[0] = EnvName; env_tuple[1] = EnvKey
    env_tuple = _get_environment_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Builds the query and arguments for the call to the API
    query = "{}/{}/{}/{}/{}/{}/status".format(ENVIRONMENTS_ENDPOINT, env_tuple[1],
                                              ENVIRONMENT_APPLICATIONS_ENDPOINT, app_tuple[1],
                                              ENVIRONMENT_APPLICATIONS_SOURCECODE_ENDPOINT, kwargs["pkg_key"])
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOURCECODE_PACKAGE_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(
            kwargs["pkg_key"], ENVIRONMENT_SOURCECODE_STATUS_FILE)
        filename = os.path.join(ENVIRONMENT_SOURCECODE_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENT_SOURCECODE_FAILED_CODE:
        raise ServerError("Failed to access the source code package status of an application. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns download link of source code package of the specified application in a given environment.
def get_environment_app_source_code_link(artifact_dir: str, endpoint: str, auth_token: str, **kwargs):
    # Tuple with (AppName, AppKey): app_tuple[0] = AppName; app_tuple[1] = AppKey
    app_tuple = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Tuple with (EnvName, EnvKey): env_tuple[0] = EnvName; env_tuple[1] = EnvKey
    env_tuple = _get_environment_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    # Builds the query and arguments for the call to the API
    query = "{}/{}/{}/{}/{}/{}/download".format(ENVIRONMENTS_ENDPOINT, env_tuple[1],
                                                ENVIRONMENT_APPLICATIONS_ENDPOINT, app_tuple[1],
                                                ENVIRONMENT_APPLICATIONS_SOURCECODE_ENDPOINT, kwargs["pkg_key"])
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOURCECODE_LINK_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(
            kwargs["pkg_key"], ENVIRONMENT_SOURCECODE_LINK_FILE)
        filename = os.path.join(ENVIRONMENT_SOURCECODE_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENT_SOURCECODE_FAILED_CODE:
        raise ServerError("Failed to access the source code package link of an application. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# ---------------------- PRIVATE METHODS ----------------------
# Private method to get the App name or key into a tuple (name,key).
def _get_environment_info(artifact_dir: str, api_url: str, auth_token: str, **kwargs):
    if "env_name" in kwargs:
        env_key = _find_environment_key(
            artifact_dir, api_url, auth_token, kwargs["env_name"])
        env_name = kwargs["env_name"]
    elif "env_key" in kwargs:
        env_key = kwargs["env_key"]
        env_name = _find_environment_name(
            artifact_dir, api_url, auth_token, kwargs["env_key"])
    else:
        raise InvalidParametersError(
            "You need to use either env_name=<name> or env_key=<key> as parameters to call this method.")
    return env_name, env_key


# Private method to find an environment key from name
def _find_environment_key(artifact_dir: str, api_url: str, auth_token: str, environment_name: str):
    env_key = ""
    cached_results = False
    try:
        # Try searching the key on the cache
        environments = load_data(artifact_dir, ENVIRONMENTS_FILE)
        cached_results = True
    except:
        # Query the LT API, since there's no cache
        environments = get_environments(artifact_dir, api_url, auth_token)
    for env in environments:
        if env["Name"] == environment_name:
            env_key = env["Key"]
            break
    # If the env key  was not found, determine if it needs to invalidate the cache or the application does not exist
    # since we explitly clear the cache, and the code is not multithreaded, it should not lead to recursion issues
    # If the cache was not used in the first place, it means the app does not exist
    if env_key == "" and not cached_results:
        raise EnvironmentNotFoundError(
            "Failed to retrieve the environment. Please make sure the environment exists. Environment name: {}".format(environment_name))
    # If the cache was used, it needs to be cleared and re-fetched from the LT server
    elif env_key == "" and cached_results:
        clear_cache(artifact_dir, ENVIRONMENTS_FILE)
        return _find_environment_key(artifact_dir, api_url, auth_token, environment_name)
    return env_key


# Private method to find an environment name from key
def _find_environment_name(artifact_dir: str, api_url: str, auth_token: str, environment_key: str):
    env_name = ""
    cached_results = False
    try:
        # Try searching the key on the cache
        environments = load_data(artifact_dir, ENVIRONMENTS_FILE)
        cached_results = True
    except:
        # Query the LT API, since there's no cache
        environments = get_environments(artifact_dir, api_url, auth_token)
    for env in environments:
        if env["Key"] == environment_key:
            env_name = env["Name"]
            break
    # If the env key  was not found, determine if it needs to invalidate the cache or the application does not exist
    # since we explitly clear the cache, and the code is not multithreaded, it should not lead to recursion issues
    # If the cache was not used in the first place, it means the app does not exist
    if env_name == "" and not cached_results:
        raise EnvironmentNotFoundError(
            "Failed to retrieve the environment. Please make sure the environment exists. Environment key: {}".format(environment_key))
    # If the cache was used, it needs to be cleared and re-fetched from the LT server
    elif env_name == "" and cached_results:
        clear_cache(artifact_dir, ENVIRONMENTS_FILE)
        return _find_environment_name(artifact_dir, api_url, auth_token, environment_key)
    return env_name


def _find_environment_url(artifact_dir: str, api_url: str, auth_token: str, environment_name: str):
    env_url = ""
    cached_results = False
    try:
        # Try searching the key on the cache
        environments = load_data(artifact_dir, ENVIRONMENTS_FILE)
        cached_results = True
    except:
        # Query the LT API, since there's no cache
        environments = get_environments(artifact_dir, api_url, auth_token)
    for env in environments:
        if env["Name"] == environment_name:
            env_url = env["HostName"]
    # If the env key  was not found, determine if it needs to invalidate the cache or the application does not exist
    # since we explitly clear the cache, and the code is not multithreaded, it should not lead to recursion issues
    # If the cache was not used in the first place, it means the app does not exist
    if env_url == "" and not cached_results:
        raise EnvironmentNotFoundError(
            "Failed to retrieve the environment. Please make sure the environment exists. Environment name: {}".format(environment_name))
    # If the cache was used, it needs to be cleared and re-fetched from the LT server
    elif env_url == "" and cached_results:
        clear_cache(artifact_dir, ENVIRONMENTS_FILE)
        return _find_environment_url(artifact_dir, api_url, auth_token, environment_name)
    return env_url



# Python Modules
import json
import os

# Custom Modules
# Exceptions
from outsystems.exceptions.no_deployments import NoDeploymentsError
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
from outsystems.exceptions.server_error import ServerError
# Functions
from outsystems.lifetime.lifetime_base import send_post_request, send_get_request
from outsystems.file_helpers.file import store_data
# Variables
from outsystems.vars.lifetime_vars import ENVIRONMENTS_ENDPOINT, ENVIRONMENT_SOLUTION_ENDPOINT, ENVIRONMENT_SOLUTION_SUCCESS_CODE, \
    ENVIRONMENT_SOLUTION_STATUS_ENDPOINT, ENVIRONMENT_SOLUTION_STATUS_SUCCESS_CODE, ENVIRONMENT_SOLUTION_STATUS_NOT_STATUS_CODE, ENVIRONMENT_SOLUTION_STATUS_NO_PERMISSION_CODE, \
    ENVIRONMENT_SOLUTION_STATUS_FAILED_CODE, ENVIRONMENT_SOLUTION_LINK_SUCCESS_CODE, ENVIRONMENT_SOLUTION_LINK_FAILED_CODE
from outsystems.vars.file_vars import SOLUTIONS_LINK_FILE, SOLUTIONS_FOLDER  # , SOLUTIONS_STATUS_FILE


# Sends a request to create a solution, on a target environment, for a specific set of app keys.
# Returns a solution key.
def create_solution(artifact_dir: str, endpoint: str, auth_token: str, environment_key: str, solution_name: str, app_keys: list, include_refs: bool):
    # Builds the API call
    query = "{}/{}/{}".format(ENVIRONMENTS_ENDPOINT, environment_key, ENVIRONMENT_SOLUTION_ENDPOINT)

    # Builds the body for the request
    solution_request = _create_solution_request(solution_name, app_keys, include_refs)
    # Sends the request
    response = send_post_request(
        endpoint, auth_token, query, solution_request)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOLUTION_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the status of a given solution key
def get_solution_status(artifact_dir: str, endpoint: str, auth_token: str, environment_key: str, solution_key: str):
    # Builds the API call
    query = "{}/{}/{}/{}".format(ENVIRONMENTS_ENDPOINT, environment_key, ENVIRONMENT_SOLUTION_STATUS_ENDPOINT, solution_key)

    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOLUTION_STATUS_SUCCESS_CODE:
        # Stores the result
        # filename = "{}{}".format(solution_key, SOLUTIONS_STATUS_FILE)
        # filename = os.path.join(SOLUTIONS_FOLDER, filename)
        # store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == ENVIRONMENT_SOLUTION_STATUS_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the details of that solution. Details: {}".format(response["response"]))
    elif status_code == ENVIRONMENT_SOLUTION_STATUS_NOT_STATUS_CODE:
        raise NoDeploymentsError("There is no solution with the key {}. Details: {}".format(
            solution_key, response["response"]))
    elif status_code == ENVIRONMENT_SOLUTION_STATUS_FAILED_CODE:
        raise ServerError("Failed to get the status of solution with key {}. Details: {}".format(
            solution_key, response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns download link of source code package of the specified application in a given environment.
def get_solution_url(artifact_dir: str, endpoint: str, auth_token: str, environment_key: str, solution_key: str):
    # Builds the API call
    query = "{}/{}/{}/{}".format(ENVIRONMENTS_ENDPOINT, environment_key, ENVIRONMENT_SOLUTION_ENDPOINT, solution_key)

    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)
    status_code = int(response["http_status"])
    if status_code == ENVIRONMENT_SOLUTION_LINK_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(solution_key, SOLUTIONS_LINK_FILE)
        filename = os.path.join(SOLUTIONS_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]["url"]
    elif status_code == ENVIRONMENT_SOLUTION_LINK_FAILED_CODE:
        raise ServerError("Failed to access the solution package link. Details: {}".format(
            response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# ---------------------- PRIVATE METHODS ----------------------
def _create_solution_request(solution_name: str, app_keys: str, include_refs: bool):

    solution_request = {"SolutionName": solution_name,
                        "ApplicationKeys": app_keys,
                        "IncludeReferences": include_refs}

    return json.dumps(solution_request)



# Base Properties API Variables
PROPERTIES_API_HTTP_PROTO = "https"
PROPERTIES_API_ENDPOINT = "PropertiesAPI/rest"
PROPERTIES_API_VERSION = 1
PROPERTIES_API_SSL_CERT_VERIFY = True

# Properties API Endpoints
SET_SITE_PROPERTY_VALUE_SUCCESS_CODE = 201
SET_REST_ENDPOINT_URL_SUCCESS_CODE = 201
SET_SOAP_ENDPOINT_URL_SUCCESS_CODE = 201
SET_TIMER_SCHEDULE_SUCCESS_CODE = 201

# Supported Property types
PROPERTY_TYPE_SITE_PROPERTY = "SiteProperty"
PROPERTY_TYPE_REST_ENDPOINT = "REST_Endpoint"
PROPERTY_TYPE_SOAP_ENDPOINT = "SOAP_Endpoint"
PROPERTY_TYPE_TIMER_SCHEDULE = "TimerSchedule"



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
from outsystems.vars.bdd_vars import BDD_HTTP_PROTO, BDD_API_ENDPOINT, BDD_CLIENT_API_ENDPOINT, BDD_API_VERSION, \
    BDD_FRAMEWORK_TYPE_CLIENT
from outsystems.vars.cicd_vars import PROBE_HTTP_PROTO, PROBE_API_ENDPOINT, PROBE_API_VERSION
from outsystems.vars.file_vars import ARTIFACT_FOLDER, BDD_FRAMEWORK_FOLDER, BDD_FRAMEWORK_TEST_ENDPOINTS_FILE
from outsystems.bdd_framework.bdd_base import build_bdd_endpoint, build_bdd_test_endpoint
from outsystems.cicd_probe.cicd_scan import scan_bdd_test_endpoint
from outsystems.cicd_probe.cicd_base import build_probe_endpoint
from outsystems.file_helpers.file import store_data, load_data
from outsystems.vars.vars_base import load_configuration_file
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_APPLICATION_NAME, MANIFEST_FLAG_IS_TEST_APPLICATION
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError
# Functions
# Variables

# ---------------------- VARS ----------------------
# Set script local variables
bdd_test = []  # will contain the BDD Framework tests for each app
bdd_modules = 0  # will count the number of bdd test modules
test_names = []  # will contain the names of the tests to run
test_list = []  # will contain the webflows output from BDD for the application
test_urls = []  # will contain the urls for the BDD framework


# ---------------------- SCRIPT ----------------------
def main(artifact_dir: str, apps: list, trigger_manifest: dict, bdd_http_proto: str, bdd_url: str, bdd_api_endpoint: str,
         bdd_client_api_endpoint: str, bdd_version: int, cicd_http_proto: str, cicd_url: str, cicd_api_endpoint: str,
         cicd_version: int, cicd_key: str, exclude_pattern: str):
    # use the script variables
    global bdd_test, bdd_modules, test_names, test_list, test_urls

    probe_endpoint = build_probe_endpoint(
        cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version)
    bdd_endpoint = build_bdd_endpoint(
        bdd_http_proto, bdd_url, bdd_api_endpoint, bdd_version)
    bdd_endpoint_client = build_bdd_endpoint(
        bdd_http_proto, bdd_url, bdd_client_api_endpoint, bdd_version)

    # Extract application list from manifest (if needed)
    if trigger_manifest:
        app_versions = trigger_manifest[MANIFEST_APPLICATION_VERSIONS]
        apps = [app_version[MANIFEST_APPLICATION_NAME] for app_version in app_versions if app_version[MANIFEST_FLAG_IS_TEST_APPLICATION]]

    # Query the CICD probe
    for app in apps:
        # Removes whitespaces in the beginning and end of the string
        app = app.strip()
        response = scan_bdd_test_endpoint(artifact_dir, probe_endpoint, app, cicd_key, exclude_pattern, cicd_version)
        if len(response) == 0:
            continue  # It has no test suites, continue the loop
        for test_endpoint in response:
            # Get the BDD test endpoints information, per module
            if cicd_version == 1 and "WebFlows" in test_endpoint["BDDTestEndpointsInfo"]:
                bdd_test += [{"EspaceName": test_endpoint["BDDTestEndpointsInfo"]["EspaceName"],
                              "WebFlows": test_endpoint["BDDTestEndpointsInfo"]["WebFlows"]}]
            elif cicd_version == 2 and "TestFlows" in test_endpoint:
                bdd_test += [{"EspaceName": test_endpoint["EspaceName"],
                              "BDDFrameworkType": test_endpoint["BDDFrameworkType"],
                              "TestFlows": test_endpoint["TestFlows"]}]
            else:
                raise NotImplementedError("Unsupported CICD Probe API version ({}).".format(cicd_version))
        # Increment bdd modules counter
        bdd_modules += len(response)

    print("{} BDD module(s) found.".format(bdd_modules), flush=True)

    # Get the tests to run (just for presentation)
    for bdd in bdd_test:  # For each BDD test
        if cicd_version == 1:
            for webflow in bdd["WebFlows"]:  # For each webflow
                if "WebScreens" in webflow:  # Sanity check to see if there are actual webscreens in the flow
                    test_list += webflow["WebScreens"]
        elif cicd_version == 2:
            for testflow in bdd["TestFlows"]:  # For each uiflow
                if "TestScreens" in testflow:  # Sanity check to see if there are actual testscreens in the flow
                    test_list += testflow["TestScreens"]

    print("{} BDD endpoint(s) scanned successfully.".format(len(test_list)), flush=True)

    # Get the names of the tests to run (just for presentation)
    for test in test_list:
        test_names.append(test["Name"])
    print("Tests to run: {}".format(test_names), flush=True)

    # For each test, generate the URL to query the BDD framework, to be used in the test class
    test_urls = []
    for bdd in bdd_test:  # For each BDD test module
        if cicd_version == 1:
            for webflow in bdd["WebFlows"]:  # For each webflow
                if "WebScreens" in webflow:  # Sanity check to see if there are actual webscreens in the flow
                    for webscreen in webflow["WebScreens"]:  # for each webscreen
                        test_endpoint = build_bdd_test_endpoint(bdd_endpoint, bdd["EspaceName"], webscreen["Name"])
                        test_urls.append(
                            {"TestSuite": bdd["EspaceName"], "Name": webscreen["Name"], "URL": test_endpoint})
        elif cicd_version == 2:
            for testflow in bdd["TestFlows"]:  # For each uiflow
                if "TestScreens" in testflow:  # Sanity check to see if there are actual testscreens in the flow
                    for testscreen in testflow["TestScreens"]:  # for each testscreen
                        if bdd["BDDFrameworkType"] == BDD_FRAMEWORK_TYPE_CLIENT:
                            target_bdd_endpoint = bdd_endpoint_client
                        else:
                            target_bdd_endpoint = bdd_endpoint
                        test_endpoint = build_bdd_test_endpoint(
                            target_bdd_endpoint, bdd["EspaceName"], testscreen["Name"])
                        test_urls.append(
                            {"TestSuite": bdd["EspaceName"], "Name": testscreen["Name"], "URL": test_endpoint})

    # Save the test results in a file for later processing
    filename = os.path.join(BDD_FRAMEWORK_FOLDER,
                            BDD_FRAMEWORK_TEST_ENDPOINTS_FILE)
    store_data(artifact_dir, filename, test_urls)


# end of main()

if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str,
                        help="Name of the artifacts folder. Default: \"Artifacts\"", default=ARTIFACT_FOLDER)
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("--cicd_probe_env", type=str,
                        help="URL for CICD Probe, without the API endpoint. Example: \"https://<host>\"", required=True)
    parser.add_argument("--cicd_probe_api", type=str,
                        help="(optional) Used to set the API endpoint for CICD Probe, without the version. Default: \"CI_CDProbe/rest\"", default=PROBE_API_ENDPOINT)
    parser.add_argument("--cicd_probe_version", type=int,
                        help="(optional) CICD Probe API version number. Default: 1", default=PROBE_API_VERSION)
    parser.add_argument("--cicd_probe_key", type=str,
                        help="(optional) Key for CICD Probe API calls (when enabled).")
    parser.add_argument("--exclude_pattern", type=str,
                        help="(optional) Regex for excluding specific ScreenFlows whose screens are not valid test endpoints")
    parser.add_argument("--bdd_framework_env", type=str,
                        help="URL for BDD Framework, without the API endpoint. Example: \"https://<host>\"", required=True)
    parser.add_argument("--bdd_framework_api", type=str,
                        help="(optional) Used to set the API endpoint for BDD Framework, without the version. Default: \"BDDFramework/rest\"", default=BDD_API_ENDPOINT)
    parser.add_argument("--bdd_framework_client_api", type=str,
                        help="(optional) Used to set the API endpoint for BDD Framework Client-side, without the version. Default: \"TestRunner_API/rest\"", default=BDD_CLIENT_API_ENDPOINT)
    parser.add_argument("--bdd_framework_version", type=int,
                        help="(optional) BDD Framework API version number. Default: 1", default=BDD_API_VERSION)
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse App list (if it exists)
    apps = None
    if args.app_list:
        apps = args.app_list.split(',')
    # Parse Manifest file (if it exists)
    manifest_file = None
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        manifest_file = load_data(manifest_path[0], manifest_path[1])
    # Check if either an app list or a manifest file is being provided
    if not args.app_list and not args.manifest_file:
        raise InvalidParametersError("either --app_list or --manifest_file must be provided as arguments")

    # Parse the BDD API endpoint
    bdd_api_endpoint = args.bdd_framework_api
    # Parse the BDD Client-side API endpoint
    bdd_client_api_endpoint = args.bdd_framework_client_api
    # Parse the BDD Url and split the BDD hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    bdd_http_proto = BDD_HTTP_PROTO
    bdd_url = args.bdd_framework_env
    if bdd_url.startswith("http://"):
        bdd_http_proto = "http"
        bdd_url = bdd_url.replace("http://", "")
    else:
        bdd_url = bdd_url.replace("https://", "")
    if bdd_url.endswith("/"):
        bdd_url = bdd_url[:-1]
    # Parse BDD API Version
    bdd_version = args.bdd_framework_version

    # Parse the CICD Probe API endpoint
    cicd_api_endpoint = args.cicd_probe_api
    # Parse the CICD Probe Url and split the CICD Probe hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    cicd_http_proto = PROBE_HTTP_PROTO
    cicd_url = args.cicd_probe_env
    if cicd_url.startswith("http://"):
        cicd_http_proto = "http"
        cicd_url = cicd_url.replace("http://", "")
    else:
        cicd_url = cicd_url.replace("https://", "")
    if cicd_url.endswith("/"):
        cicd_url = cicd_url[:-1]
    # Parse CICD Probe API Version
    cicd_version = args.cicd_probe_version
    # Parse CICD Probe API Key
    cicd_key = args.cicd_probe_key
    # Parse Exclude Pattern regex
    exclude_pattern = args.exclude_pattern

    # Calls the main script
    main(artifact_dir, apps, manifest_file, bdd_http_proto, bdd_url, bdd_api_endpoint, bdd_client_api_endpoint, bdd_version,
         cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version, cicd_key, exclude_pattern)



# Python Modules
import sys
import os
import argparse
from time import sleep
import json

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, SOLUTIONS_OSP_FILE, SOLUTIONS_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_APPLICATION_KEY, MANIFEST_FLAG_IS_TEST_APPLICATION, MANIFEST_APPLICATION_NAME
from outsystems.vars.pipeline_vars import SOLUTION_TIMEOUT_IN_SECS, SOLUTION_SLEEP_PERIOD_IN_SECS, SOLUTION_CREATED_STATUS, \
    SOLUTION_READY_STATUS, SOLUTION_GATHERING_DEPENDENCIES_STATUS, SOLUTION_GETTING_BINARIES_STATUS, SOLUTION_GENERATING_META_MODEL_STATUS, \
    SOLUTION_GENERATING_SOLUTION_STATUS, SOLUTION_COMPLETED_STATUS, SOLUTION_ABORTED_STATUS

# Functions
from outsystems.file_helpers.file import load_data, bytes_human_readable_size
from outsystems.lifetime.lifetime_solutions import create_solution, get_solution_status, get_solution_url
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.lifetime.lifetime_downloads import download_package
from outsystems.manifest.manifest_base import get_environment_details
from outsystems.vars.vars_base import get_configuration_value, load_configuration_file
# Exceptions
from outsystems.exceptions.manifest_does_not_exist import ManifestDoesNotExistError


# ############################################################# SCRIPT ##############################################################
# Get a formatted status message for the different statuses of a solution generation process.
def get_status_message(status: str):
    status_messages = {
        SOLUTION_CREATED_STATUS: "An empty solution was created in the system - it does not contain any associated application.",
        SOLUTION_READY_STATUS: "All included applications were added to the database and the solution is ready to be processed.",
        SOLUTION_GATHERING_DEPENDENCIES_STATUS: "Calculating all the dependencies for the solution.",
        SOLUTION_GETTING_BINARIES_STATUS: "Getting the binaries for each module included in the solution.",
        SOLUTION_GENERATING_META_MODEL_STATUS: "Building the solution package manifest.",
        SOLUTION_GENERATING_SOLUTION_STATUS: "Creating the solution file."
    }

    return status_messages.get(status, "Unknown status: " + status)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env_label: str, include_test_apps: bool, solution_name: str, include_refs: bool, trigger_manifest: dict):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Tuple with (EnvName, EnvKey): src_env_tuple[0] = EnvName; src_env_tuple[1] = EnvKey
    env_tuple = get_environment_details(trigger_manifest, source_env_label)

    # Retrive the app keys from the manifest content
    application_keys = [app[MANIFEST_APPLICATION_KEY] for app in trigger_manifest.get(MANIFEST_APPLICATION_VERSIONS, []) if include_test_apps or not app.get(MANIFEST_FLAG_IS_TEST_APPLICATION)]

    # Send request to create a solution with the given app keys
    solution_key = create_solution(artifact_dir, lt_endpoint, lt_token, env_tuple[1], solution_name, application_keys, include_refs)

    # Wait for solution package creation to finish
    wait_counter = 0
    package_url_available = False
    check_status = None
    IN_PROGESS_STATUS = [SOLUTION_CREATED_STATUS, SOLUTION_READY_STATUS, SOLUTION_GATHERING_DEPENDENCIES_STATUS,
                         SOLUTION_GETTING_BINARIES_STATUS, SOLUTION_GENERATING_META_MODEL_STATUS,
                         SOLUTION_GENERATING_SOLUTION_STATUS]

    # Retrieve the app names from the manifest content
    application_names = [app[MANIFEST_APPLICATION_NAME] for app in trigger_manifest.get(MANIFEST_APPLICATION_VERSIONS, []) if include_test_apps or not app.get(MANIFEST_FLAG_IS_TEST_APPLICATION)]

    # Print information about the solution package
    print("A solution package will be created from '{}', containing the latest version of each module from the following applications:".format(env_tuple[0]), flush=True)

    for app_name in application_names:
        print(" - {} ".format(app_name), flush=True)

    # Print additional information if include_refs is True
    if include_refs:
        print("Producer modules will also be included in the solution package", flush=True)

    print("Start creation of '{}' package:".format(solution_name), flush=True)
    while wait_counter < get_configuration_value("SOLUTION_TIMEOUT_IN_SECS", SOLUTION_TIMEOUT_IN_SECS):
        # Check current package status
        solution_status = get_solution_status(artifact_dir, lt_endpoint, lt_token, env_tuple[1], solution_key)
        if solution_status["Status"] == SOLUTION_COMPLETED_STATUS:
            # Package was created successfully
            package_url_available = True
            break
        elif solution_status["Status"] == SOLUTION_ABORTED_STATUS:
            print(" - {}. Reason: {}".format(solution_status["Status"], solution_status["StatusReason"]), flush=True)
            exit(1)
        elif solution_status["Status"] in IN_PROGESS_STATUS:
            # Solution package is still being created. Go back to sleep.
            sleep_value = get_configuration_value("SOLUTION_SLEEP_PERIOD_IN_SECS", SOLUTION_SLEEP_PERIOD_IN_SECS)
            sleep(sleep_value)
            wait_counter += sleep_value
            if check_status != solution_status["Status"]:
                print(" - {} - {}".format(solution_status["Status"], get_status_message(solution_status["Status"])), flush=True)
                check_status = solution_status["Status"]
        else:
            raise NotImplementedError("Unknown solution code status: {}.".format(solution_status["Status"]))

    # When the package is created, download it using the provided key
    if package_url_available:
        print("Solution package {} created successfully.".format(solution_key), flush=True)
        solution_url = get_solution_url(artifact_dir, lt_endpoint, lt_token, env_tuple[1], solution_key)

        file_name = solution_name + SOLUTIONS_OSP_FILE
        file_path = os.path.join(artifact_dir, SOLUTIONS_FOLDER, file_name)
        download_package(file_path, lt_token, solution_url)

        print("Solution package successfully downloaded as '{}' ({}).".format(file_name, bytes_human_readable_size(os.path.getsize(file_path))), flush=True)
    else:
        print("Timeout expired while generating solution package {}.".format(solution_key), flush=True)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="(Optional) Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="(Optional) LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env_label", type=str, required=True,
                        help="Label, as configured in the manifest, of the source environment where the apps are.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="Flag that indicates if applications marked as \"Test Application\" in the manifest are included in the solution.")
    parser.add_argument("-m", "--trigger_manifest", type=str,
                        help="Manifest artifact (in JSON format) received when the pipeline is triggered. Contains required data used throughout the pipeline execution.")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file (with JSON format). Contains required data used throughout the pipeline execution.")
    parser.add_argument("-sn", "--solution_name", type=str, required=True,
                        help="Name of the solution package that will be created.")
    parser.add_argument("-r", "--include_refs", action='store_true',
                        help="Flag that indicates if whether to include producer modules in the solution.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env_label = args.source_env_label
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps
    # Parse Solution Name
    solution_name = args.solution_name
    # Parse Include References flag
    include_refs = args.include_refs

    # Validate Manifest is being passed either as JSON or as file
    if not args.trigger_manifest and not args.manifest_file:
        raise ManifestDoesNotExistError("The manifest was not provided as JSON or as a file. Aborting!")

    # Parse Trigger Manifest artifact
    if args.manifest_file:
        trigger_manifest_path = os.path.split(args.manifest_file)
        trigger_manifest = load_data(trigger_manifest_path[0], trigger_manifest_path[1])
    else:
        trigger_manifest = json.loads(args.trigger_manifest)

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env_label, include_test_apps, solution_name, include_refs, trigger_manifest)



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, APPLICATION_OAP_FOLDER, APPLICATION_OAP_FILE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION, DEPLOYMENT_MESSAGE
from outsystems.vars.cicd_vars import PROBE_HTTP_PROTO, PROBE_API_ENDPOINT, PROBE_API_VERSION
# Functions
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.lifetime.lifetime_applications import export_app_oap
from outsystems.file_helpers.file import load_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.cicd_probe.cicd_base import build_probe_endpoint
from outsystems.osp_tool.osp_base import call_osptool
from outsystems.cicd_probe.cicd_dependencies import get_app_dependencies, sort_app_dependencies
from outsystems.pipeline.deploy_latest_tags_to_target_env import generate_deployment_based_on_manifest as generate_deployment_based_on_deploy_manifest, \
    generate_regular_deployment
from outsystems.pipeline.deploy_tags_to_target_env_with_manifest import generate_deployment_based_on_manifest as generate_deployment_based_on_trigger_manifest
from outsystems.vars.vars_base import load_configuration_file
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError


# ############################################################# SCRIPT ##############################################################
def generate_oap_list(app_data_list: list, friendly_package_names: bool):
    app_oap_list = []
    filename = ""
    for app in app_data_list:
        if friendly_package_names:
            filename = "{}_v{}{}".format(app["Name"].replace(" ", "_"), app["Version"].replace(".", "_"), APPLICATION_OAP_FILE)
        else:
            filename = "{}{}".format(app["VersionKey"], APPLICATION_OAP_FILE)
        app_oap_list.append({"app_name": app["Name"], "app_version": app["Version"], "app_key": app["Key"], "version_key": app["VersionKey"], "filename": filename})
    return app_oap_list


def export_apps_oap(artifact_dir: str, lt_endpoint: str, lt_token: str, env_key: str, app_oap_list: list):
    print("Application Scope:", flush=True)
    for app in app_oap_list:
        file_path = os.path.join(artifact_dir, APPLICATION_OAP_FOLDER, app["filename"])
        export_app_oap(file_path, lt_endpoint, lt_token, env_key, app_key=app["app_key"], app_version_key=app["version_key"])
        print("     {} application with version {}, exported as {}".format(app["app_name"], app["app_version"], app["filename"]), flush=True)


def generate_deployment_order(artifact_dir: str, probe_endpoint: str, api_key: str, app_oap_list: list):
    dependencies_list = {}
    for app in app_oap_list:
        dependencies_list[app["app_key"]] = get_app_dependencies(artifact_dir, probe_endpoint, api_key,
                                                                 app["version_key"], app["app_name"], app["app_version"])
    dependencies_order_list = sort_app_dependencies(dependencies_list)
    final_list = []
    for app_dep in dependencies_order_list:
        for app_oap in app_oap_list:
            if app_dep == app_oap["app_key"]:
                final_list.append(app_oap)
    return final_list


def deploy_apps_oap(artifact_dir: str, dest_env: str, osp_tool_path: str, credentials: str, app_oap_list: list):
    for app in app_oap_list:
        oap_file_path = os.path.join(artifact_dir, APPLICATION_OAP_FOLDER, app["filename"])
        call_osptool(osp_tool_path, oap_file_path, dest_env, credentials)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env: str, dest_env: str, apps: list, dep_manifest: list, trigger_manifest: dict, include_test_apps: bool, dep_note: str, osp_tool_path: str, credentials: str, cicd_http_proto: str, cicd_url: str, cicd_api_endpoint: str, cicd_version: str, cicd_key: str, friendly_package_names: bool):

    app_data_list = []  # will contain the applications to deploy details from LT

    if lt_api_version == 1:  # LT for OS version < 11
        raise InvalidParametersError("Air Gap deployments are not supported for Deployment API v1")

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)
    # Builds the Probe endpoint
    probe_endpoint = build_probe_endpoint(cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version)

    # Gets the environment key for the source environment
    src_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, source_env)

    # If the manifest file is being used, the app versions MUST come from that file
    # Or else you might not be deploying the same app versions that were deployed in
    # previous pipeline steps
    if dep_manifest:
        app_data_list = generate_deployment_based_on_deploy_manifest(artifact_dir, lt_endpoint, lt_token, src_env_key, source_env, apps, dep_manifest)
    elif trigger_manifest:
        app_data_list = generate_deployment_based_on_trigger_manifest(artifact_dir, lt_endpoint, lt_token, src_env_key, source_env, trigger_manifest, include_test_apps, False)
    else:
        app_data_list = generate_regular_deployment(artifact_dir, lt_endpoint, lt_token, src_env_key, apps)

    # Export binary files
    app_oap_list = generate_oap_list(app_data_list, friendly_package_names)
    export_apps_oap(artifact_dir, lt_endpoint, lt_token, src_env_key, app_oap_list)

    # Generate deployment order
    sorted_oap_list = generate_deployment_order(artifact_dir, probe_endpoint, cicd_key, app_oap_list)

    print("\nDeployment Order:\n", flush=True)
    for oap in sorted_oap_list:
        print("      " + str(sorted_oap_list.index(oap) + 1) + ". " + oap["app_name"] + " (" + oap["version_key"] + ")\n", flush=True)

    # Deploy binary files to target environment
    deploy_apps_oap(artifact_dir, dest_env, osp_tool_path, credentials, sorted_oap_list)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the source environment where the apps are.")
    parser.add_argument("-d", "--destination_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the destination environment where you want to deploy the apps. (if in Airgap mode should be the hostname of the destination environment where you want to deploy the apps)")
    parser.add_argument("-m", "--deploy_msg", type=str, default=DEPLOYMENT_MESSAGE,
                        help="Message you want to show on the deployment plans in LifeTime. Default: \"Automated deploy using OS Pipelines\".")
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file path (either deployment or trigger), used to promote the same application versions throughout the pipeline execution.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="(Optional) Flag that indicates if applications marked as \"Test Application\" in the trigger manifest are included in the Air Gap deployment.")
    parser.add_argument("-o", "--osp_tool_path", type=str, required=True,
                        help="OSP Tool file path")
    parser.add_argument("-user", "--airgap_user", type=str, required=True,
                        help="Username with privileges to deploy applications on target environment")
    parser.add_argument("-pass", "--airgap_pass", type=str, required=True,
                        help="Password of the Username with priveleges to deploy applications on target environment")
    parser.add_argument("-pu", "--cicd_probe_url", type=str, required=True,
                        help="URL of the environment where the CI/CD Probe is installed (without the API endpoint).")
    parser.add_argument("-pv", "--cicd_probe_version", type=str, default=PROBE_API_VERSION,
                        help="(Optional) CI/CD Probe API version number.")
    parser.add_argument("-pe", "--cicd_probe_endpoint", type=str, default=PROBE_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for CI/CD Probe, without the version.")
    parser.add_argument("-pk", "--cicd_probe_key", type=str,
                        help="(Optional) Key for CI/CD Probe API calls (when enabled).")
    parser.add_argument("-n", "--friendly_package_names", action='store_true',
                        help="Flag that indicates if downloaded application packages should have a user-friendly name. Example: \"AppName_v1_2_1\"")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env = args.source_env
    # Parse Destination Environment
    dest_env = args.destination_env
    # Parse Manifest file if it exists
    # Based on the file content it can be a deployment manifest (list-based) or trigger manifest (dict-based)
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        manifest_file = load_data(manifest_path[0], manifest_path[1])
    else:
        manifest_file = None
    dep_manifest = manifest_file if type(manifest_file) is list else None
    trigger_manifest = manifest_file if type(manifest_file) is dict else None
    # Check if either an app list or a manifest file is being provided
    if not args.app_list and not args.manifest_file:
        raise InvalidParametersError("either --app_list or --manifest_file must be provided as arguments")

    if dep_manifest and not args.app_list:
        raise InvalidParametersError("--app_list parameter is required for Deployment Manifest operation")
    # Parse App list
    apps = None
    if args.app_list:
        _apps = args.app_list
        apps = _apps.split(',')
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps
    # Parse Deployment Message
    dep_note = args.deploy_msg
    # Parse OSP Tool path
    osp_tool_path = args.osp_tool_path
    # Parse Credentials for OSP Tool
    credentials = args.airgap_user + " " + args.airgap_pass
    # Parse the CICD Probe API endpoint
    cicd_api_endpoint = args.cicd_probe_endpoint
    # Parse the CICD Probe Url and split the CICD Probe hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    cicd_http_proto = PROBE_HTTP_PROTO
    cicd_url = args.cicd_probe_url
    if cicd_url.startswith("http://"):
        cicd_http_proto = "http"
        cicd_url = cicd_url.replace("http://", "")
    else:
        cicd_url = cicd_url.replace("https://", "")
    if cicd_url.endswith("/"):
        cicd_url = cicd_url[:-1]
    # Parse CICD Probe API Version
    cicd_version = args.cicd_probe_version
    # Parse CICD Probe API Key
    cicd_key = args.cicd_probe_key
    # Parse Friendly Package Names flag
    friendly_package_names = args.friendly_package_names

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env, dest_env, apps, dep_manifest, trigger_manifest, include_test_apps, dep_note, osp_tool_path, credentials, cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version, cicd_key, friendly_package_names)



import subprocess
import threading

# Custom Modules
# Variables
from outsystems.vars.pipeline_vars import SOLUTION_TIMEOUT_IN_SECS
# Functions
from outsystems.vars.vars_base import get_configuration_value
# Exceptions
from outsystems.exceptions.osptool_error import OSPToolDeploymentError


def run_command(command, live_output_callback=None, timeout=None):
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

    def read_output(pipe, callback, output_list):
        with pipe:
            for line in iter(pipe.readline, ''):
                callback(line.strip())
                output_list.append(line)

    # List to capture the live output
    live_output = []

    # Create a thread for reading and displaying live output
    live_output_thread = threading.Thread(target=read_output, args=(process.stdout, live_output_callback, live_output))
    live_output_thread.start()

    # Wait for the process to finish and get the return code
    try:
        return_code = process.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        # Process has exceeded the timeout
        process.terminate()
        raise OSPToolDeploymentError("OSPTool Deployment timed out.")

    # Wait for the live output thread to finish
    live_output_thread.join()

    # Combine the live output into a single string (execution log)
    execution_log = ''.join(live_output)

    return return_code, execution_log


def call_osptool(osp_tool_path: str, package_file_path: str, env_hostname: str, credentials: str, catalogmappings_path: str):

    if catalogmappings_path:
        # Construct the command using a formatted string
        command = '"{}" "{}" "{}" {} /catalogmappings "{}"'.format(osp_tool_path, package_file_path, env_hostname, credentials, catalogmappings_path)
    else:
        # Construct the command using a formatted string
        command = '"{}" "{}" "{}" {}'.format(osp_tool_path, package_file_path, env_hostname, credentials)

    # Define a callback function for live output
    def live_output_callback(output_line):
        print(output_line)

    # Run the command and get the return code and execution log
    return_code, execution_log = run_command(command, live_output_callback, timeout=get_configuration_value("SOLUTION_TIMEOUT_IN_SECS", SOLUTION_TIMEOUT_IN_SECS))

    return return_code, execution_log



from distutils.core import setup
import os

NAME = 'outsystems-pipeline'
DESCRIPTION = 'Python package to accelerate the integration of OutSystems with third-party CI/CD tools'
LONG_DESCRIPTION = '''The outsystems-pipeline Python package provides functions to support the creation of OutSystems CI/CD pipelines using your DevOps automation tool of choice.

Visit the `project repository <https://github.com/OutSystems/outsystems-pipeline>`_ on GitHub for instructions on how to build example OutSystems CI/CD pipelines with common DevOps automation tools, as well as documentation that will help you adapt the examples to your particular scenarios.


What's new
==========

**Solution Package Deployments**

 Fixed input parameter conflict on the following script:

 * `deploy_package_to_target_env.py`

**Updated Package Dependencies**

 * Updated `requests` dependency to version 2.32.5
 * Updated python version requirement (>=3.9)

Installing and upgrading
========================

Install or upgrade outsystems-pipeline to the latest available version as follows:
::

    pip install -U outsystems-pipeline

'''
AUTHOR = u'OutSystems'
EMAIL = u'cicd.integrations@outsystems.com'
URL = 'https://github.com/OutSystems/outsystems-pipeline'
LICENSE = 'Apache License 2.0'
PYTHON_REQUIRES = '>=3.9'
KEYWORDS = [
    '',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'License :: OSI Approved :: Apache Software License',
    'Programming Language :: Python',
    'Topic :: Software Development :: Build Tools',
    'Topic :: Software Development :: Quality Assurance',
    'Topic :: Software Development :: Testing',
    'Topic :: Software Development :: Testing :: Acceptance',
    'Topic :: Software Development :: Testing :: BDD',
    'Topic :: Software Development :: Testing :: Unit',
    'Topic :: System :: Software Distribution'
]

REQUIREMENTS = [
    'python-dateutil==2.9.0.post0',
    'requests==2.32.5',
    'unittest-xml-reporting==3.2.0',
    'xunitparser==1.3.4',
    'toposort==1.10',
    'python-dotenv==1.0.1',
    'packaging==24.1'
]

PACKAGES = [
    'outsystems',
    'outsystems.architecture_dashboard',
    'outsystems.bdd_framework',
    'outsystems.cicd_probe',
    'outsystems.exceptions',
    'outsystems.file_helpers',
    'outsystems.lifetime',
    'outsystems.manifest',
    'outsystems.osp_tool',
    'outsystems.pipeline',
    'outsystems.properties',
    'outsystems.vars'
]

if __name__ == '__main__':  # Do not run setup() when we import this module.
    if os.path.isfile("VERSION"):
        with open("VERSION", 'r') as version_file:
            version = version_file.read().replace('\n', '')
    else:
        # dummy version
        version = '1.0.0'

    setup(
        name=NAME,
        version='0.10.2',
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        keywords=' '.join(KEYWORDS),
        author=AUTHOR,
        author_email=EMAIL,
        url=URL,
        license=LICENSE,
        python_requires=PYTHON_REQUIRES,
        classifiers=CLASSIFIERS,
        packages=PACKAGES,
        install_requires=REQUIREMENTS
    )



# Python Modules

# Custom Modules
# Functions
from outsystems.properties.properties_base import send_properties_put_request
# Variables
from outsystems.vars.properties_vars import SET_SITE_PROPERTY_VALUE_SUCCESS_CODE, SET_TIMER_SCHEDULE_SUCCESS_CODE, \
    SET_REST_ENDPOINT_URL_SUCCESS_CODE, SET_SOAP_ENDPOINT_URL_SUCCESS_CODE


def set_site_property_value(lt_url: str, token: str, module_key: str, environment_key: str, site_property_key: str, site_property_value: str):
    # Builds the API endpoint
    api_endpoint = "Modules/{}/Environments/{}/SiteProperties/{}/Value/".format(module_key, environment_key, site_property_key)
    # Sends the request
    response = send_properties_put_request(
        lt_url, token, api_endpoint, site_property_value)
    status_code = response["http_status"]
    if status_code == SET_SITE_PROPERTY_VALUE_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


def set_rest_endpoint_url(lt_url: str, token: str, module_key: str, environment_key: str, rest_endpoint_key: str, rest_endpoint_url: str):
    # Builds the API endpoint
    api_endpoint = "Modules/{}/Environments/{}/RESTReferences/{}/EffectiveURL/".format(module_key, environment_key, rest_endpoint_key)
    # Sends the request
    response = send_properties_put_request(
        lt_url, token, api_endpoint, rest_endpoint_url)
    status_code = response["http_status"]
    if status_code == SET_REST_ENDPOINT_URL_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


def set_soap_endpoint_url(lt_url: str, token: str, module_key: str, environment_key: str, soap_endpoint_key: str, soap_endpoint_url: str):
    # Builds the API endpoint
    api_endpoint = "Modules/{}/Environments/{}/SOAPReferences/{}/EffectiveURL/".format(module_key, environment_key, soap_endpoint_key)
    # Sends the request
    response = send_properties_put_request(
        lt_url, token, api_endpoint, soap_endpoint_url)
    status_code = response["http_status"]
    if status_code == SET_SOAP_ENDPOINT_URL_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


def set_timer_schedule(lt_url: str, token: str, module_key: str, environment_key: str, timer_key: str, timer_schedule: str):
    # Builds the API endpoint
    api_endpoint = "Modules/{}/Environments/{}/Timers/{}/Schedule/".format(module_key, environment_key, timer_key)
    # Sends the request
    response = send_properties_put_request(
        lt_url, token, api_endpoint, timer_schedule)
    status_code = response["http_status"]
    if status_code == SET_TIMER_SCHEDULE_SUCCESS_CODE:
        return response["response"]
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))



# Python Modules
import sys
import os
import argparse

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION, DEPLOYMENT_MESSAGE
from outsystems.vars.manifest_vars import MANIFEST_APPLICATION_VERSIONS, MANIFEST_ENVIRONMENT_DEFINITIONS, MANIFEST_DEPLOYMENT_NOTES, \
    MANIFEST_APPLICATION_KEY, MANIFEST_APPLICATION_NAME, MANIFEST_APPLICATION_VERSION_KEY, MANIFEST_APPLICATION_VERSION_NUMBER, \
    MANIFEST_ENVIRONMENT_KEY, MANIFEST_ENVIRONMENT_NAME, MANIFEST_ENVIRONMENT_LABEL, MANIFEST_FLAG_IS_TEST_APPLICATION, \
    MANIFEST_FOLDER, MANIFEST_FILE

# Functions
from outsystems.lifetime.lifetime_environments import get_environments, get_environment_deployment_zones
from outsystems.lifetime.lifetime_applications import get_running_app_version, get_application_data
from outsystems.file_helpers.file import store_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint


# Function that will build the info required for the environments
def generate_manifest_env_info(artifact_dir: str, lt_endpoint: str, lt_token: str):
    # Gets all infra environments information
    infra_envs = get_environments(artifact_dir, lt_endpoint, lt_token)

    # Trims info to include only the desired env info (Name and Key)
    env_info = [{MANIFEST_ENVIRONMENT_KEY: env["Key"], MANIFEST_ENVIRONMENT_NAME: env["Name"], MANIFEST_ENVIRONMENT_LABEL: env["Name"]} for env in infra_envs if "Name" in env and "Key" in env]

    return env_info


# Function that will build the info required for a deployment based on the latest versions of the apps in the src environment
def generate_manifest_app_info(artifact_dir: str, lt_endpoint: str, lt_token: str, src_env_key: str, app_list: list):
    app_data_list = []  # will contain the applications to deploy details from LT

    deployment_zones = get_environment_deployment_zones(artifact_dir, lt_endpoint, lt_token, env_key=src_env_key)

    # Creates a list with the details for the apps you want to deploy
    for app_name in app_list:
        # Removes whitespaces in the beginning and end of the string
        app_name = app_name.strip()

        # Get the app running version on the source environment. It will only retrieve tagged applications
        app_info = get_running_app_version(artifact_dir, lt_endpoint, lt_token, src_env_key, app_name=app_name)

        # Get the module info
        app_module_data = get_application_data(artifact_dir, lt_endpoint, lt_token, True, app_key=app_info[MANIFEST_APPLICATION_KEY])

        # Get deployment zone info
        deployment_zone_key = next((item['DeploymentZoneKey'] for item in app_module_data['AppStatusInEnvs'] if item['EnvironmentKey'] == src_env_key), None)
        deployment_zone_name = next((item['Name'] for item in deployment_zones if item['Key'] == deployment_zone_key), None)

        # Add it to the app data list
        app_data_list.append({MANIFEST_APPLICATION_KEY: app_info[MANIFEST_APPLICATION_KEY], MANIFEST_APPLICATION_NAME: app_info[MANIFEST_APPLICATION_NAME],
                              MANIFEST_APPLICATION_VERSION_KEY: app_info[MANIFEST_APPLICATION_VERSION_KEY], MANIFEST_APPLICATION_VERSION_NUMBER: app_info["Version"],
                              'CreatedOn': app_info["CreatedOn"], 'ChangeLog': app_info["ChangeLog"], MANIFEST_FLAG_IS_TEST_APPLICATION: False,
                              'DeploymentZoneKey': deployment_zone_key, 'DeploymentZoneName': deployment_zone_name})

    return app_data_list


# Function that will generate and save the manifest file
def generate_manifest_file(artifact_dir: str, app_details: list, env_details: list, dep_note: str):

    manifest_data = {
        MANIFEST_APPLICATION_VERSIONS: app_details,
        MANIFEST_ENVIRONMENT_DEFINITIONS: env_details,
        MANIFEST_DEPLOYMENT_NOTES: dep_note
    }

    # Store the manifest to be used in other stages of the pipeline
    filename = "{}/{}".format(MANIFEST_FOLDER, MANIFEST_FILE)
    store_data(artifact_dir, filename, manifest_data)
    return manifest_data


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env: str, apps: list, dep_note: str):

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Save environments info structure for manifest
    env_details = generate_manifest_env_info(artifact_dir, lt_endpoint, lt_token)

    # Gets the environment key for the source environment
    src_env_key = next((env["EnvironmentKey"] for env in env_details if env.get("EnvironmentName") == source_env), None)

    # Save applications info structure for manifest
    app_details = generate_manifest_app_info(artifact_dir, lt_endpoint, lt_token, src_env_key, apps)

    generate_manifest_file(artifact_dir, app_details, env_details, dep_note)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the source environment where the apps are.")
    parser.add_argument("-l", "--app_list", type=str, required=True,
                        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-m", "--deploy_msg", type=str, default=DEPLOYMENT_MESSAGE,
                        help="Message you want to show on the deployment plans in LifeTime. Default: \"Automated deploy using OS Pipelines\".")
    args = parser.parse_args()

    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env = args.source_env
    # Parse App list
    _apps = args.app_list
    apps = _apps.split(',')
    # Parse Deployment Message
    dep_note = args.deploy_msg

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env, apps, dep_note)



# Python Modules
import requests

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_json_response import InvalidJsonResponseError


# Method that builds the endpoint for Architecture Dashboard API and returns it
def build_ad_endpoint(ad_http_proto: str, ad_api_host: str, ad_api_endpoint: str, ad_api_version: int):
    return "{}://{}/{}/v{}".format(ad_http_proto, ad_api_host, ad_api_endpoint, ad_api_version)


# Sends a GET request to Architecture Dashboard
def send_get_request(request_string: str, activation_code: str, api_key: str, url_params: dict = None):
    # API key + Customer Activation Code
    headers = {'x-api-key': api_key,
               'x-activation-code': activation_code}

    response = requests.get(request_string, params=url_params, headers=headers)
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "GET {}: The JSON response could not be parsed. Response: {}".format(request_string, response.text))
    return response_obj



# Python Modules
import requests
import json

# Custom Modules
# Exceptions
from outsystems.exceptions.invalid_json_response import InvalidJsonResponseError
# Variables
from outsystems.vars.lifetime_vars import LIFETIME_SSL_CERT_VERIFY
# Functions
from outsystems.vars.vars_base import get_configuration_value
from outsystems.file_helpers.file import check_file


# Method that builds the LifeTime endpoint based on the LT host
def build_lt_endpoint(lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int):
    # Builds the endpoint for LT and returns it
    return "{}://{}/{}/v{}".format(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)


# Sends a GET request to LT, with url_params
def send_get_request(lt_api: str, token: str, api_endpoint: str, url_params: dict):
    # Auth token + content type json
    headers = {'content-type': 'application/json',
               'authorization': 'Bearer ' + token}
    # Format the request URL to include the api endpoint
    request_string = "{}/{}".format(lt_api, api_endpoint)
    response = requests.get(request_string, params=url_params, headers=headers, verify=get_configuration_value("LIFETIME_SSL_CERT_VERIFY", LIFETIME_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "GET {}: The JSON response could not be parsed. Response: {}".format(request_string, response.text))
    return response_obj


# Sends a POST request to LT, with a payload. The json part is ignored
def send_post_request(lt_api: str, token: str, api_endpoint: str, payload: str):
    # Auth token + content type json
    headers = {'content-type': 'application/json',
               'authorization': 'Bearer ' + token}
    # Format the request URL to include the api endpoint
    request_string = "{}/{}".format(lt_api, api_endpoint)
    response = requests.post(
        request_string, data=payload, json=None, headers=headers, verify=get_configuration_value("LIFETIME_SSL_CERT_VERIFY", LIFETIME_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    # Since LT API POST requests do not reply with native JSON, we have to make it ourselves
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            # Workaround for POST /deployments/ since the response is not JSON, just text
            response_obj["response"] = json.loads('"{}"'.format(response.text))
    return response_obj


# Sends a POST request to LT, with binary content.
def send_binary_post_request(lt_api: str, token: str, api_endpoint: str, dest_env: str, lt_endpont: str, binary_file_path: str):
    # Auth token + content type octet-stream
    headers = {'content-type': 'application/octet-stream',
               'authorization': 'Bearer ' + token}
    # Format the request URL to include the api endpoint
    request_string = "{}/{}/{}/{}".format(lt_api, api_endpoint, dest_env, lt_endpont)

    if check_file("", binary_file_path):
        with open(binary_file_path, 'rb') as f:
            data = f.read()
    response = requests.post(request_string, data=data, headers=headers, verify=get_configuration_value("LIFETIME_SSL_CERT_VERIFY", LIFETIME_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    # Since LT API POST requests do not reply with native JSON, we have to make it ourselves
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            # Workaround for POST /deployments/ since the response is not JSON, just text
            response_obj["response"] = json.loads('"{}"'.format(response.text))
    return response_obj


# Sends a DELETE request to LT
def send_delete_request(lt_api: str, token: str, api_endpoint: str):
    # Auth token + content type json
    headers = {'content-type': 'application/json',
               'authorization': 'Bearer ' + token}
    # Format the request URL to include the api endpoint
    request_string = "{}/{}".format(lt_api, api_endpoint)
    response = requests.delete(request_string, headers=headers, verify=get_configuration_value("LIFETIME_SSL_CERT_VERIFY", LIFETIME_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": {}}
    if len(response.text) > 0:
        try:
            response_obj["response"] = response.json()
        except:
            raise InvalidJsonResponseError(
                "DELETE {}: The JSON response could not be parsed. Response: {}".format(request_string, response.text))
    return response_obj


# Sends a GET request to LT, with url_params
def send_download_request(pkg_url: str, token: str):
    # Auth token + content type json
    headers = {'content-type': 'application/json',
               'authorization': token}
    # Format the request URL to include the api endpoint
    response = requests.get(pkg_url, headers=headers, verify=get_configuration_value("LIFETIME_SSL_CERT_VERIFY", LIFETIME_SSL_CERT_VERIFY))
    response_obj = {"http_status": response.status_code, "response": response.content}
    return response_obj



# Python Modules
import os
import json

# Custom Modules
# Exceptions
from outsystems.exceptions.app_does_not_exist import AppDoesNotExistError
from outsystems.exceptions.invalid_parameters import InvalidParametersError
from outsystems.exceptions.no_apps_available import NoAppsAvailableError
from outsystems.exceptions.server_error import ServerError
from outsystems.exceptions.not_enough_permissions import NotEnoughPermissionsError
from outsystems.exceptions.environment_not_found import EnvironmentNotFoundError
from outsystems.exceptions.app_version_error import AppVersionsError
# Functions
from outsystems.file_helpers.file import store_data, load_data, clear_cache
from outsystems.lifetime.lifetime_base import send_get_request, send_post_request
from outsystems.lifetime.lifetime_downloads import download_package
# Variables
from outsystems.vars.file_vars import APPLICATION_FOLDER, APPLICATIONS_FILE, APPLICATION_FILE, APPLICATION_VERSIONS_FILE, APPLICATION_VERSION_FILE
from outsystems.vars.lifetime_vars import APPLICATIONS_ENDPOINT, APPLICATION_VERSIONS_ENDPOINT, APPLICATIONS_SUCCESS_CODE, \
    APPLICATIONS_EMPTY_CODE, APPLICATIONS_FLAG_FAILED_CODE, APPLICATIONS_FAILED_CODE, APPLICATION_SUCCESS_CODE, \
    APPLICATION_FLAG_FAILED_CODE, APPLICATION_NO_PERMISSION_CODE, APPLICATION_FAILED_CODE, APPLICATION_VERSION_SUCCESS_CODE, \
    APPLICATION_VERSION_INVALID_CODE, APPLICATION_VERSION_NO_PERMISSION_CODE, APPLICATION_VERSION_FAILED_CODE, \
    APPLICATION_VERSION_FAILED_LIST_CODE, APPLICATION_VERSIONS_CONTENT, APPLICATION_VERSIONS_EMPTY_CODE, \
    ENVIRONMENTS_ENDPOINT, ENVIRONMENT_APPLICATIONS_ENDPOINT, APPLICATION_VERSION_CREATE_SUCCESS_CODE, APPLICATION_VERSION_CREATE_INVALID_CODE, \
    APPLICATION_VERSION_CREATE_NO_PERMISSION_CODE, APPLICATION_VERSION_CREATE_NO_ENVIRONMENT_CODE, APPLICATION_VERSION_CREATE_FAILED_CODE


# Returns a list of applications that exist in the infrastructure.
def get_applications(artifact_dir: str, endpoint: str, auth_token: str, extra_data: bool):
    params = {"IncludeModules": extra_data, "IncludeEnvStatus": extra_data}
    # Sends the request
    response = send_get_request(
        endpoint, auth_token, APPLICATIONS_ENDPOINT, params)
    status_code = int(response["http_status"])
    # Process the response based on the status code returned from the server
    if status_code == APPLICATIONS_SUCCESS_CODE:
        # Stores the result
        store_data(artifact_dir, APPLICATIONS_FILE, response["response"])
        return response["response"]
    elif status_code == APPLICATIONS_EMPTY_CODE:
        raise NoAppsAvailableError(
            "No applications available in the infrastructure. Details {}".format(response["response"]))
    elif status_code == APPLICATIONS_FLAG_FAILED_CODE:
        raise InvalidParametersError(
            "There was an error with the 'extra_data' flag or the request was invalid when listing all applications. The params used were: {}. Details: {}".format(params, response[
                "response"]))
    elif status_code == APPLICATIONS_FAILED_CODE:
        raise ServerError(
            "Failed to list the applications. Details {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns the details of a given application.
def get_application_data(artifact_dir: str, endpoint: str, auth_token: str, extra_data: bool, **kwargs):
    # Tuple with (AppName, AppKey): app_info[0] = AppName; app_info[1] = AppKey
    app_info = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    query = "{}/{}".format(APPLICATIONS_ENDPOINT, app_info[1])
    params = {"IncludeModules": extra_data, "IncludeEnvStatus": extra_data}
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, params)
    status_code = int(response["http_status"])
    if status_code == APPLICATION_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(app_info[0], APPLICATION_FILE)
        filename = os.path.join(APPLICATION_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == APPLICATION_FLAG_FAILED_CODE:
        raise InvalidParametersError(
            "There was an error with the 'extra_data' flag or the request was invalid when listing the application. The params used were: {}. Details: {}".format(params, response[
                "response"]))
    elif status_code == APPLICATION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the details of that application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_FAILED_CODE:
        raise EnvironmentNotFoundError(
            "Failed getting running applications because one of the environments was not found. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Returns a list of versions of a given application.
def get_application_versions(artifact_dir: str, endpoint: str, auth_token: str, number_of_versions: int, **kwargs):
    # Tuple with (AppName, AppKey): app_info[0] = AppName; app_info[1] = AppKey
    app_info = _get_application_info(
        artifact_dir, endpoint, auth_token, **kwargs)
    query = "{}/{}/{}".format(APPLICATIONS_ENDPOINT,
                              app_info[1], APPLICATION_VERSIONS_ENDPOINT)
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, {"MaximumVersionsToReturn": number_of_versions})
    status_code = int(response["http_status"])
    if status_code == APPLICATION_VERSION_SUCCESS_CODE:
        # Stores the result
        filename = "{}{}".format(app_info[0], APPLICATION_VERSIONS_FILE)
        filename = os.path.join(APPLICATION_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == APPLICATION_VERSION_INVALID_CODE:
        raise InvalidParametersError(
            "Invalid request due to invalid max versions to return (less than 0). Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the versions of that application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_CODE:
        raise AppDoesNotExistError(
            "Failed to retrieve the application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_LIST_CODE:
        raise AppVersionsError(
            "Failed to list the application versions. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


def get_application_version(artifact_dir: str, endpoint: str, auth_token: str, extra_data: bool, version_id: str, **kwargs):
    # Tuple with (AppName, AppKey): app_info[0] = AppName; app_info[1] = AppKey
    app_info = _get_application_info(artifact_dir, endpoint, auth_token, **kwargs)
    query = "{}/{}/{}/{}".format(APPLICATIONS_ENDPOINT,
                                 app_info[1], APPLICATION_VERSIONS_ENDPOINT, version_id)
    # Sends the request
    params = {"IncludeModules": extra_data, "IncludeEnvStatus": extra_data}
    response = send_get_request(endpoint, auth_token, query, params)
    status_code = int(response["http_status"])
    if status_code == APPLICATION_VERSION_SUCCESS_CODE:
        # Stores the result
        filename = "{}.{}{}".format(app_info[0], version_id, APPLICATION_VERSION_FILE)
        filename = os.path.join(APPLICATION_FOLDER, filename)
        store_data(artifact_dir, filename, response["response"])
        return response["response"]
    elif status_code == APPLICATION_VERSION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the versions of that application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_CODE:
        raise AppDoesNotExistError(
            "Failed to retrieve the application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_LIST_CODE:
        raise AppVersionsError(
            "Failed to retrieve the application version. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


def get_running_app_version(artifact_dir: str, endpoint: str, auth_token: str, env_key: str, **kwargs):
    # Tuple with (AppName, AppKey): app_tuple[0] = AppName; app_tuple[1] = AppKey
    app_tuple = _get_application_info(artifact_dir, endpoint, auth_token, **kwargs)
    app_data = {}

    deployed_app = get_application_data(artifact_dir, endpoint, auth_token, True, app_name=app_tuple[0])
    for status_in_env in deployed_app["AppStatusInEnvs"]:
        if status_in_env["EnvironmentKey"] == env_key:
            app_version_data = get_application_version(artifact_dir, endpoint, auth_token, True, status_in_env["BaseApplicationVersionKey"], app_name=app_tuple[0])
            app_data = {
                "ApplicationName": app_tuple[0],
                "ApplicationKey": app_tuple[1],
                "Version": app_version_data["Version"],
                "VersionKey": status_in_env["BaseApplicationVersionKey"],
                "IsModified": status_in_env["IsModified"]
            }
            # Since these 2 fields were only introduced in a minor of OS11, we check here if they exist
            # We can't just use the version
            if "CreatedOn" in app_version_data:
                app_data.update({"CreatedOn": app_version_data["CreatedOn"]})
            if "ChangeLog" in app_version_data:
                app_data.update({"ChangeLog": app_version_data["ChangeLog"]})
            break

    return app_data


def set_application_version(endpoint: str, auth_token: str, env_key: str, app_key: str, change_log: str, app_version: str, mobile_versions: list):
    query = "{}/{}/{}/{}/{}".format(ENVIRONMENTS_ENDPOINT,
                                    env_key, ENVIRONMENT_APPLICATIONS_ENDPOINT, app_key, APPLICATION_VERSIONS_ENDPOINT)

    version_request = {"ChangeLog": change_log,
                       "Version": app_version, "MobileVersions": mobile_versions}

    response = send_post_request(endpoint, auth_token, query, json.dumps(version_request))

    status_code = int(response["http_status"])
    if status_code == APPLICATION_VERSION_CREATE_SUCCESS_CODE:
        return response["response"]
    elif status_code == APPLICATION_VERSION_CREATE_INVALID_CODE:
        raise InvalidParametersError("The request is invalid. Check the body of the request for errors. Body: {}. Details: {}.".format(
            version_request, response["response"]))
    elif status_code == APPLICATION_VERSION_CREATE_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to create the version. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_CREATE_NO_ENVIRONMENT_CODE:
        raise EnvironmentNotFoundError(
            "Can't find the application or target environment. Details: {}.".format(response["response"]))
    elif status_code == APPLICATION_VERSION_CREATE_FAILED_CODE:
        raise ServerError(
            "Failed to tag an application, or Failed to create a new version. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# Exports the OAP of a given application version.
def export_app_oap(file_path: str, endpoint: str, auth_token: str, env_key: str, app_key: str, app_version_key: str):
    query = "{}/{}/{}/{}/{}".format(APPLICATIONS_ENDPOINT,
                                    app_key, APPLICATION_VERSIONS_ENDPOINT, app_version_key, APPLICATION_VERSIONS_CONTENT)
    # Sends the request
    response = send_get_request(endpoint, auth_token, query, None)

    status_code = int(response["http_status"])
    if status_code == APPLICATIONS_SUCCESS_CODE:
        # Stores the result
        url_string = response["response"]
        url_string = url_string["url"]
        download_package(file_path, auth_token, url_string)
        return
    elif status_code == APPLICATION_VERSION_NO_PERMISSION_CODE:
        raise NotEnoughPermissionsError(
            "You don't have enough permissions to see the details of that application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_INVALID_CODE:
        raise AppVersionsError(
            "The request is invalid for the given keys. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSIONS_EMPTY_CODE:
        raise AppDoesNotExistError(
            "No binary available for given keys. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_CODE:
        raise EnvironmentNotFoundError(
            "Failed to retrieve the application. Details: {}".format(response["response"]))
    elif status_code == APPLICATION_VERSION_FAILED_LIST_CODE:
        raise ServerError(
            "Failed to download the oap of the application version. Details: {}".format(response["response"]))
    else:
        raise NotImplementedError(
            "There was an error. Response from server: {}".format(response))


# ---------------------- PRIVATE METHODS ----------------------

# Private method to get the App name or key into a tuple (name,key).
def _get_application_info(artifact_dir: str, api_url: str, auth_token: str, **kwargs):
    if "app_name" in kwargs:
        app_key = _find_application_key(
            artifact_dir, api_url, auth_token, kwargs["app_name"])
        app_name = kwargs["app_name"]
    elif "app_key" in kwargs:
        app_key = kwargs["app_key"]
        app_name = _find_application_name(
            artifact_dir, api_url, auth_token, kwargs["app_key"])
    else:
        raise InvalidParametersError(
            "You need to use either app_name=<name> or app_key=<key> as parameters to call this method.")
    return (app_name, app_key)


# Private method to find an application key from name
def _find_application_key(artifact_dir: str, api_url: str, auth_token: str, application_name: str):
    app_key = ""
    cached_results = False
    try:
        # Try searching the key on the cache
        applications = load_data(artifact_dir, APPLICATIONS_FILE)
        cached_results = True
    except FileNotFoundError:
        # Query the LT API, since there's no cache
        applications = get_applications(
            artifact_dir, api_url, auth_token, False)
    for app in applications:
        if app["Name"] == application_name:
            app_key = app["Key"]
            break
    # If the app key was not found, determine if it needs to invalidate the cache or the application does not exist
    # since we explitly clear the cache, and the code is not multithreaded, it should not lead to recursion issues
    # If the cache was not used in the first place, it means the app does not exist
    if app_key == "" and not cached_results:
        raise AppDoesNotExistError(
            "Failed to retrieve the application. Please make sure the app exists in the environment. App Name: {}".format(application_name))
    # If the cache was used, it needs to be cleared and re-fetched from the LT server
    elif app_key == "" and cached_results:
        clear_cache(artifact_dir, APPLICATIONS_FILE)
        return _find_application_key(artifact_dir, api_url, auth_token, application_name)
    return app_key


# Private method to find an application name from key
def _find_application_name(artifact_dir: str, api_url: str, auth_token: str, application_key: str):
    app_name = ""
    cached_results = False
    try:
        # Try searching the key on the cache
        applications = load_data(artifact_dir, APPLICATIONS_FILE)
        cached_results = True
    except FileNotFoundError:
        # Query the LT API, since there's no cache
        applications = get_applications(
            artifact_dir, api_url, auth_token, False)
    for app in applications:
        if app["Key"] == application_key:
            app_name = app["Name"]
            break
    # If the app name  was not found, determine if it needs to invalidate the cache or the application does not exist
    # since we explitly clear the cache, and the code is not multithreaded, it should not lead to recursion issues
    # If the cache was not used in the first place, it means the app does not exist
    if app_name == "" and not cached_results:
        raise AppDoesNotExistError(
            "Failed to retrieve the application. Please make sure the app exists in the environment. App Key: {}".format(application_key))
    # If the cache was used, it needs to be cleared and re-fetched from the LT server
    elif app_name == "" and cached_results:
        clear_cache(artifact_dir, APPLICATIONS_FILE)
        return _find_application_name(artifact_dir, api_url, auth_token, application_key)
    return app_name



# Python Modules
import sys
import os
import argparse
import json

# Workaround for Jenkins:
# Set the path to include the outsystems module
# Jenkins exposes the workspace directory through env.
if "WORKSPACE" in os.environ:
    sys.path.append(os.environ['WORKSPACE'])
else:  # Else just add the project dir
    sys.path.append(os.getcwd())

# Custom Modules
# Variables
from outsystems.vars.file_vars import ARTIFACT_FOLDER, APPLICATION_OAP_FOLDER, DEPLOYMENT_ORDER_FILE
from outsystems.vars.lifetime_vars import LIFETIME_HTTP_PROTO, LIFETIME_API_ENDPOINT, LIFETIME_API_VERSION
from outsystems.vars.cicd_vars import PROBE_HTTP_PROTO, PROBE_API_ENDPOINT, PROBE_API_VERSION

# Functions
from outsystems.lifetime.lifetime_environments import get_environment_key
from outsystems.file_helpers.file import load_data
from outsystems.lifetime.lifetime_base import build_lt_endpoint
from outsystems.pipeline.deploy_latest_tags_to_target_env import generate_deployment_based_on_manifest as generate_deployment_based_on_deploy_manifest, generate_regular_deployment
from outsystems.pipeline.deploy_tags_to_target_env_with_manifest import generate_deployment_based_on_manifest as generate_deployment_based_on_trigger_manifest
from outsystems.pipeline.deploy_apps_to_target_env_with_airgap import export_apps_oap, generate_deployment_order, generate_oap_list
from outsystems.cicd_probe.cicd_base import build_probe_endpoint
from outsystems.vars.vars_base import load_configuration_file
# Exceptions
from outsystems.exceptions.invalid_parameters import InvalidParametersError


# ################################## SCRIPT ##################################
def store_data(filename: str, data: list):
    with open(filename, 'w') as f:
        json.dump(data, f)


def main(artifact_dir: str, lt_http_proto: str, lt_url: str, lt_api_endpoint: str, lt_api_version: int, lt_token: str, source_env: str, apps: list, dep_manifest: list, trigger_manifest: dict, include_test_apps: bool, cicd_http_proto: str, cicd_url: str, cicd_api_endpoint: str, cicd_version: str, cicd_key: str, friendly_package_names: bool, generate_deploy_order: bool):

    # will contain the applications to deploy details from LT
    app_data_list = []

    if lt_api_version == 1:  # LT for OS version < 11
        raise InvalidParametersError("Not supported for LifeTime API v1")

    # Builds the LifeTime endpoint
    lt_endpoint = build_lt_endpoint(lt_http_proto, lt_url, lt_api_endpoint, lt_api_version)

    # Gets the environment key for the source environment
    src_env_key = get_environment_key(artifact_dir, lt_endpoint, lt_token, source_env)

    # If the manifest file is being used, the app versions MUST come from that file
    # Or else you might not be deploying the same app versions that were deployed in
    # previous pipeline steps
    if dep_manifest:
        app_data_list = generate_deployment_based_on_deploy_manifest(artifact_dir, lt_endpoint, lt_token, src_env_key, source_env, apps, dep_manifest)
    elif trigger_manifest:
        app_data_list = generate_deployment_based_on_trigger_manifest(artifact_dir, lt_endpoint, lt_token, src_env_key, source_env, trigger_manifest, include_test_apps, False)
    else:
        app_data_list = generate_regular_deployment(artifact_dir, lt_endpoint, lt_token, src_env_key, apps)

    # Export binary files
    app_oap_list = generate_oap_list(app_data_list, friendly_package_names)
    export_apps_oap(artifact_dir, lt_endpoint, lt_token, src_env_key, app_oap_list)

    if generate_deploy_order:
        # Builds the Probe endpoint
        probe_endpoint = build_probe_endpoint(cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version)

        # Generate deployment order
        sorted_oap_list = generate_deployment_order(artifact_dir, probe_endpoint, cicd_key, app_oap_list)

        print("\nDeployment Order:\n", flush=True)
        for oap in sorted_oap_list:
            print("      " + str(sorted_oap_list.index(oap) + 1) + ". " + oap["app_name"] + " (" + oap["version_key"] + ")\n", flush=True)

        # Export Deployment Order as File
        store_data(os.path.join(artifact_dir, APPLICATION_OAP_FOLDER, DEPLOYMENT_ORDER_FILE), sorted_oap_list)

# End of main()


if __name__ == "__main__":
    # Argument menu / parsing
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--artifacts", type=str, default=ARTIFACT_FOLDER,
                        help="Name of the artifacts folder. Default: \"Artifacts\"")
    parser.add_argument("-u", "--lt_url", type=str, required=True,
                        help="URL for LifeTime environment, without the API endpoint. Example: \"https://<lifetime_host>\"")
    parser.add_argument("-t", "--lt_token", type=str, required=True,
                        help="Token for LifeTime API calls.")
    parser.add_argument("-v", "--lt_api_version", type=int, default=LIFETIME_API_VERSION,
                        help="LifeTime API version number. If version <= 10, use 1, if version >= 11, use 2. Default: 2")
    parser.add_argument("-e", "--lt_endpoint", type=str, default=LIFETIME_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for LifeTime, without the version. Default: \"lifetimeapi/rest\"")
    parser.add_argument("-s", "--source_env", type=str, required=True,
                        help="Name, as displayed in LifeTime, of the source environment where the apps are.")
    parser.add_argument("-l", "--app_list", type=str,
                        help="Comma separated list of apps you want to deploy. Example: \"App1,App2 With Spaces,App3_With_Underscores\"")
    parser.add_argument("-f", "--manifest_file", type=str,
                        help="Manifest file path (either deployment or trigger), used to promote the same application versions throughout the pipeline execution.")
    parser.add_argument("-i", "--include_test_apps", action='store_true',
                        help="(Optional) Flag that indicates if applications marked as \"Test Application\" in the trigger manifest are included in the Air Gap deployment.")
    parser.add_argument("-pu", "--cicd_probe_url", type=str,
                        help="(Optional) URL of the environment where the CI/CD Probe is installed (without the API endpoint).")
    parser.add_argument("-pv", "--cicd_probe_version", type=str, default=PROBE_API_VERSION,
                        help="(Optional) CI/CD Probe API version number.")
    parser.add_argument("-pe", "--cicd_probe_endpoint", type=str, default=PROBE_API_ENDPOINT,
                        help="(Optional) Used to set the API endpoint for CI/CD Probe, without the version.")
    parser.add_argument("-pk", "--cicd_probe_key", type=str,
                        help="(Optional) Key for CI/CD Probe API calls (when enabled).")
    parser.add_argument("-n", "--friendly_package_names", action='store_true',
                        help="Flag that indicates if downloaded application packages should have a user-friendly name. Example: \"AppName_v1_2_1\"")
    parser.add_argument("-g", "--generate_deploy_order", action='store_true',
                        help="Flag that indicates if the deploy order file should be created.")
    parser.add_argument("-cf", "--config_file", type=str,
                        help="Config file path. Contains configuration values to override the default ones.")

    args = parser.parse_args()

    # Load config file if exists
    if args.config_file:
        load_configuration_file(args.config_file)
    # Parse the artifact directory
    artifact_dir = args.artifacts
    # Parse the API endpoint
    lt_api_endpoint = args.lt_endpoint
    # Parse the LT Url and split the LT hostname from the HTTP protocol
    # Assumes the default HTTP protocol = https
    lt_http_proto = LIFETIME_HTTP_PROTO
    lt_url = args.lt_url
    if lt_url.startswith("http://"):
        lt_http_proto = "http"
        lt_url = lt_url.replace("http://", "")
    else:
        lt_url = lt_url.replace("https://", "")
    if lt_url.endswith("/"):
        lt_url = lt_url[:-1]
    # Parte LT API Version
    lt_version = args.lt_api_version
    # Parse the LT Token
    lt_token = args.lt_token
    # Parse Source Environment
    source_env = args.source_env
    # Check if either an app list or a manifest file is being provided
    if not args.app_list and not args.manifest_file:
        raise InvalidParametersError("either --app_list or --manifest_file must be provided as arguments")
    # Parse App list
    if args.app_list:
        _apps = args.app_list
        apps = _apps.split(',')
    else:
        apps = None
    # Parse Manifest file if it exists
    # Based on the file content it can be a deployment manifest (list-based)
    # or trigger manifest (dict-based)
    if args.manifest_file:
        manifest_path = os.path.split(args.manifest_file)
        manifest_file = load_data(manifest_path[0], manifest_path[1])
    else:
        manifest_file = None
    dep_manifest = manifest_file if type(manifest_file) is list else None
    trigger_manifest = manifest_file if type(manifest_file) is dict else None
    # Parse Include Test Apps flag
    include_test_apps = args.include_test_apps
    # Parse the CICD Probe API endpoint
    cicd_api_endpoint = args.cicd_probe_endpoint
    # Parse the CICD Probe Url and split the CICD Probe hostname from the HTTP protocol
    # Assumes the default HTTP protocol = "https"
    cicd_http_proto = PROBE_HTTP_PROTO
    if args.cicd_probe_url:
        cicd_url = args.cicd_probe_url
        if cicd_url.startswith("http://"):
            cicd_http_proto = "http"
            cicd_url = cicd_url.replace("http://", "")
        else:
            cicd_url = cicd_url.replace("https://", "")
        if cicd_url.endswith("/"):
            cicd_url = cicd_url[:-1]
    else:
        cicd_url = None
    # Parse CICD Probe API Version
    cicd_version = args.cicd_probe_version
    # Parse CICD Probe API Key
    cicd_key = args.cicd_probe_key
    # Parse Friendly Package Names flag
    friendly_package_names = args.friendly_package_names
    # Parse Generate Deployment Order flag
    generate_deploy_order = args.generate_deploy_order
    if generate_deploy_order and not args.cicd_probe_url:
        raise InvalidParametersError("The CI/CD Probe is required to create the deployment order and must be provided as argument")

    # Calls the main script
    main(artifact_dir, lt_http_proto, lt_url, lt_api_endpoint, lt_version, lt_token, source_env, apps, dep_manifest, trigger_manifest, include_test_apps, cicd_http_proto, cicd_url, cicd_api_endpoint, cicd_version, cicd_key, friendly_package_names, generate_deploy_order)
